package ircbot;

import java.net.*;
import java.io.*;
import java.nio.charset.*;
import java.util.*;
import ircbot.modules.*;

/**
 * This is the main engine for the bot. This class provides methods to
 * handle the network connection, and the main loop that reads
 * messages from the IRC network. Most functionality of the bot has
 * been placed in separate modules so there isn't really that much
 * action in this class. This engine only handles one module, all
 * other modules are handled by the module <code>Modules</code>.
 */
// TODO: New ideas: Sayings, Smurf, Yatzy, Poker, Tänk till tusen.
public class IRCBot
{
    /**
     * Bot information
     */
    private IRCBotConfig botConfig;
    private Person me;

    /**
     * IRC network information
     */
    private String IRC_network;
    private int IRC_port;
    private BufferedReader IRC_in;
    private BufferedWriter IRC_out;

    /**
     * System data structures
     */
    private Modules modules;
    private Channels mod_channels;
    private People mod_people;
    private ArrayList<IRCEvent> eventlist;

    /**
     * Create a new bot given a name and a description.
     */
    public IRCBot(IRCBotConfig config)
    {
        botConfig = config;
        eventlist = new ArrayList<IRCEvent>();

        /* To keep the bot fully operational, do not remove any of the
         * modules Modules, Channels, or People.
         */
        modules = new Modules(this);
        modules.addModule(mod_channels = new Channels());
        modules.addModule(mod_people = new People());

        me = mod_people.getPerson(config.getBotName(), true);
        mod_people.getPerson(config.getOperatorNick(), true).setOperator(true);

        modules.addModule(new AI());
        modules.addModule(new Calc());
        modules.addModule(new Events());
        modules.addModule(new Help());
        modules.addModule(new Lang());
        modules.addModule(new MasterMind());
        modules.addModule(new News());
        modules.addModule(new Numbers());
        modules.addModule(new Pending());
        modules.addModule(new Quotes());
        modules.addModule(new StenSaxPase());
        modules.addModule(new Time());
        modules.addModule(new Trivia());
        modules.addModule(new Wiki());
    }

    public String getName()
    {
        return botConfig.getBotName();
    }


    /***********************************************************************
     * Event management
     */

    public void scheduleEvent(IRCEvent event)
    {
        eventlist.add(event);
        Collections.sort(eventlist);
    }

    public void unscheduleEvent(IRCEvent event)
    {
        eventlist.remove(event);
    }

    public void unscheduleEvents(IRCChannel channel)
    {
        Iterator<IRCEvent> it = eventlist.iterator();
        while (it.hasNext()) {
            IRCEvent event = it.next();
            if (event.getChannel() == channel)
                it.remove();
        }
    }

    private IRCEvent getScheduledEvent()
    {
        if (eventlist.size() > 0 &&
            eventlist.get(0).getActivationTime() < System.currentTimeMillis())
            return eventlist.remove(0);
        else
            return null;
    }

    public Collection<IRCEvent> getEvents()
    {
        return eventlist;
    }


    /***********************************************************************
     * Network connection
     */

    // TODO: Be online in more than one IRC network.
    public void connect() throws IOException
    {
        if (botConfig.getConsole())
            connect(new IRCReader(botConfig.getOperatorNick(), "#console"),
                    new BufferedWriter(new OutputStreamWriter(System.out)));
        else
            connect(botConfig.getIrcNetwork());
    }

    /**
     * Used to connect to an IRC network.
     * @param network The domain name of the network to connect to.
     * @throws IOException If the bot is unable to connect to the IRC
     * network.
     */
    public void connect(String network) throws IOException
    {
        connect(network, 6667);
    }

    /**
     * Used to connect to an IRC network.
     * @param network The domain name of the network to connect to.
     * @param port The port to connect to.
     * @throws IOException If the bot is unable to connect to the IRC
     * network.
     */
    public void connect(String network, int port) throws IOException
    {
        IRC_network = network;
        IRC_port = port;

        Charset charset = Charset.forName("ISO-8859-1");
        Socket socket = new Socket(IRC_network, IRC_port);
        IRC_in = new BufferedReader(new InputStreamReader(socket.getInputStream(), charset.newDecoder()));
        IRC_out = new BufferedWriter(new OutputStreamWriter(socket.getOutputStream(), charset.newEncoder()));

        ircSend("user " + botConfig.getBotName() + " ware2 irc :" + botConfig.getBotInfoText());
        ircSend("nick " + botConfig.getBotName());
        if (botConfig.getBotNickPassword() != null) {
            ircSend("privmsg nickserv : identify " + botConfig.getBotNickPassword());
        }

        mod_channels.joinOnline();
    }

    /**
     * Connect to the given readers. Can be used to connect to an
     * input and an output file or to stdin and stdout.
     * @param in The BufferedReader to connect to.
     * @param out The BufferedWriter to connect to.
     */
    public void connect(BufferedReader in, BufferedWriter out)
    {
        IRC_in = in;
        IRC_out = out;

        ircSend("*** You are connected ***");
    }


    /**
     * Can be used to join channels in the IRC network. Any calls to
     * this method will be passed to the Channels module.
     * @param channel The channel to join.
     * @throws IOException If the message to join the channel is not
     * sent correctly to the IRC network.
     */
    /*
    public void join(IRCChannel channel) throws IOException
    {
        ircSend(mod_channels.join(channel));
    }
    */

    /**
     * Called to terminate the network connection.
     */
    public void quit() throws IOException
    {
        ircSend("quit IaintGonnaTAKEthisAnyMoreandthatsFINAL.");
        IRC_in.close();
        IRC_out.close();
    }

    private void reconnect() throws IOException
    {
        IRC_in.close();
        IRC_out.close();

        Charset charset = Charset.forName("ISO-8859-1");
        Socket socket = new Socket(IRC_network, IRC_port);
        IRC_in = new BufferedReader(new InputStreamReader(socket.getInputStream(), charset.newDecoder()));
        IRC_out = new BufferedWriter(new OutputStreamWriter(socket.getOutputStream(), charset.newEncoder()));

        ircSend("user " + botConfig.getBotName() + " ware2 irc :" + botConfig.getBotInfoText());
        ircSend("nick " + botConfig.getBotName());
        if (botConfig.getBotNickPassword() != null) {
            ircSend("privmsg nickserv : identify " + botConfig.getBotNickPassword());
        }

        mod_channels.joinOnline();
    }

    /**
     * Called to send a correctly formatted message to the IRC
     * network. Messages that do not follow the IRC RFC will be
     * ignored by the network.
     * @param message The message to send.
     */
    private void ircSend(String message)
    {
        try {
            IRC_out.write(message);
            IRC_out.newLine();
            IRC_out.flush();
        }
        catch(IOException e) {
            System.err.println("[ircSend: " + e.getClass() + "] " + e.getMessage());
            try {
                reconnect();
            }
            catch(IOException ioe) {
                System.err.println("[ircSend: " + ioe.getClass() + "] " + ioe.getMessage());
            }
        }
    }

    /**
     * Called to send messages to the IRC network. Each message will
     * be formatted to follow the IRC RFC which pretty much means
     * adding the keyword PRIVMSG and the receiver channel or nick
     * plus making sure the message is not too long. If a message
     * starts with the string "syscmd:" no formatting will be
     * enforced.
     * @param nick The receiver of this message, channel or nick.
     * @param messages The messages to send. Use a line break character
     * ('\n') to separate messages.
     */
    private void sendMsg(String nick, String messages)
    {
        String lastPart;
        String[] message = messages.split("\n");

        for (int i = 0; i < message.length; i++) {
            if (message[i].trim().length() == 0)
                continue;

            logMessage("[SEND] " + message[i]);

            if (message[i].startsWith("syscmd:")) {
                ircSend(message[i].substring(7));
                continue;
            }

            do {
                lastPart = null;
                message[i] = "privmsg " + nick + " :" + message[i];

                if (message[i].length() > 450) {
                    int space = message[i].lastIndexOf(' ', 450);

                    if (space == -1)
                        space = 450;

                    lastPart = message[i].substring(space + 1);
                    message[i] = message[i].substring(0, space);
                }

                ircSend(message[i]);
                sleep(100);

                message[i] = lastPart;
            } while (lastPart != null);
        }
    }

    /**
     * Called to send a delayed message to the IRC network. A random
     * delay for up to three seconds is activated before the message is sent.
     * @param nick The receiver channel or nick.
     * @param message The message to send.
     * @param delayed TRUE to activate the delay, FALSE otherwise.
     */
    private void sendMsg(String nick, String message, boolean delayed)
    {
        if (delayed)
            sleep((int)Math.random() * 3000);
        sendMsg(nick, message);
    }

    private void sendMsg(IRCChannel channel, String message)
    {
        channel.channelActive();
        channel.setLastReply(message);
        sendMsg(channel.getName(), message);
    }


    /**********************************************************************
     * Main loop
     */
    public void doTheBotThing() throws IOException
    {
        boolean onLine = true;

        while (onLine) {
            if (!IRC_in.ready()) {
                /* No message was sent to the channel right now. Take
                 * a nap and then see if we can pester the channel
                 * with random thoughts.
                 */
                sleep(100);

                /* Save the settings every now and then. */
                String reply = modules.periodicSave();
                if (reply != null)
                    logMessage(reply);

                for (IRCChannel channel : mod_channels.getChannels()) {
                    if (channel.isInactive()) {
                        // TODO: On-line settings for random thought timeout.
                        channel.setLastReply(null);
                        if (channel.getRandomMessage()) {
                            reply = modules.randomThought(channel, me);
                            if (reply != null) {
                                sendMsg(channel, reply);
                            }
                        }
                    }
                }

                IRCEvent event = getScheduledEvent();
                if (event != null) {
                    String result = event.activateEvent();
                    if (result != null)
                        sendMsg(event.getChannel(), result);
                }

                continue;
            }

            String msg = IRC_in.readLine();

            /* Handle ping */
            if (msg.substring(0, 4).equalsIgnoreCase("ping")) {
                ircSend("pong " + msg.substring(5));
                continue;
            }

            /* ERROR :Closing Link: host.ip.or.name (Ping timeout: 182 seconds) */
            if (msg.startsWith("ERROR :")) {
                logMessage("[ERROR] '" + msg + "' Reconnecting...");
                reconnect();
                continue;
            }

            /* We ignore some crappy messages received from the server... */
            if (msg.charAt(0) != ':') {
                logMessage("[IGNORE] " + msg);
                continue;
            }

            /* Create an incoming message. */
            IRCMessage message = new IRCMessage(msg, mod_channels, mod_people, botConfig.getBotName());

            IRCChannel channel = message.getChannel();
            if (channel != null)
                channel.channelActive();

            logMessage("[RECEIVE] " + msg);

            /* Some people do not deserve our attention. */
            if (mod_people.ignores(message.getSender().getNick()) ||
                mod_people.isRemoved(message.getSender().getNick()))
                continue;

            /* If we get a message containing a command, handle it
             * here.
             */
            if (message.getBotCommand() != null) {
                // TODO:  Remove 'botname:' before parsing so that commands are recognized.
                String reply = modules.action(message);
                if (reply != null) {
                    if (channel != null)
                        reply = channel.filterReply(reply);

                    if (reply.equals("syscmd:quit"))
                        onLine = false;
                    else
                        sendMsg(message.getReplyTo(), reply);
                }
                continue;
            }

            /* Call parseMessage for all messages in the channel
             * (except for commands that was handled above). If we are
             * allowed to talk in this channel and a reply is
             * available, send it!
             */
            String reply = modules.parseMessage(message);

            if (channel != null)
                reply = channel.filterReply(reply);

            if (reply != null && (channel == null || channel.isVerbose())) {
                sendMsg(message.getReplyTo(), reply, true);
            }
        }

        modules.saveState();
    }


    /**********************************************************************
     * Static methods 
     */

    /**
     * This method defines what characters are considered to be
     * command prefixes. Only messages starting with one of these
     * characters will be considered commands.
     * @param ch The character to test.
     * @return TRUE if the given character was a command prefix, FALSE
     * otherwise.
     */
    public static boolean isCommandPrefix(char ch)
    {
        switch (ch) {
        case '!':
        //case '?':
        //case '+':
        //case '-':
            return true;
        }
        return false;
    }

    /**
     * Put the bot to sleep for a while.
     * @param millis The number of milliseconds to sleep.
     */
    public static void sleep(int millis)
    {
        try {
            Thread.sleep(millis);
        }
        catch(InterruptedException e) { }
    }

    /**
     * Creates a string representation of a list with all elements
     * from the given collection of Strings.
     * @param c The collection to create a list from.
     * @return A string containing the list.
     */
    public static String makeList(Collection<?>c)
    {
        ArrayList<String> list = new ArrayList<String>();

        for (Object o : c)
            list.add(o.toString());

        return makeList(list, true);
    }

    /**
     * Creates a string representation of a list with all elements
     * from the given collection of Strings. If requested, the list is
     * sorted alphabetically.
     * @param c The collection to create a list from.
     * @param sorted TRUE to generate a sorted list, FALSE otherwise.
     * @return A string containing the list.
     */
    public static String makeList(Collection<String> c, boolean sorted)
    {
        Collection<String> list;

        if (c.size() == 0)
            return "[ ]";

        if (sorted) {
            list = new TreeSet<String>(String.CASE_INSENSITIVE_ORDER);
            list.addAll(c);
        }
        else
            list = c;

        String text = "[ ";
        for (String item : list) {
            text += item + " | ";
        }
        return text.substring(0, text.length() - 2) + "]";
    }

    /**
     * Used to filter messages if a reply every single time something
     * happens is not desired.
     * @param prob The probability of a reply. A value between 0 and 1
     * where 0 means not a chance and 1 means every single time.
     * @param message The message that should be sent if we are lucky.
     * @return The message, or null if no message is to be sent.
     */
    public static String maybeReply(double prob, String message)
    {
        if (Math.random() < prob)
            return message;
        else
            return null;
    }

    public void logMessage(String msg)
    {
        if (botConfig.getLogging() == IRCBotConfig.LoggingType.STANDARD)
            System.out.println(msg);
        else if (botConfig.getLogging() == IRCBotConfig.LoggingType.COLOR) // TODO: Color logging
            System.out.println(msg);
    }
}
