package ircbot;

import java.util.*;
import ircbot.modules.*;

/**
 * This interface is used to create modules for the bot. Any module
 * that wants to be notified with traffic from the IRC network needs
 * to implement this interface or extend the AbstractIRCModule class.
 * @see AbstractIRCBotModule
 */

public interface IRCBotModule extends Comparable<IRCBotModule>
{
    /**
     * getModuleName is called to get the name of this module. The
     * name is used when building the list of loaded modules. A module
     * name may contain capital letters.
     * @return A String containing the name of this module.
     */
    public String getModuleName();

    /**
     * getCommands is called to get the command line names of all
     * commands provided by this module.  An IRC message starting with
     * a member of getCommands() will invoke this modules action
     * method. All commands are lower case and begins with the command
     * prefix (!). Note that a module's action
     * method is ONLY called for commands returned by getCommands.
     * @param type The needed privilege level needed to access this
     * command.
     * @return A Collection of String containing the commands
     * provided by this module.
     * @see AbstractIRCBotModule, Modules.CommandType
     */
    public Collection<String> getCommands(Modules.CommandType type);

    /**
     * action is invoked by the bot when someone types a message starting
     * with a member of getCommands().
     * @param message The original message sent to the
     * channel. Provided if some command needs further information
     * from it, like for instance the nick of the sender.
     * @return A String containing a message to be sent to the IRC channel
     * as a reply to the command invocation or null if no reply is needed.
     */
    public String action(IRCMessage message);

    /**
     * parseMessage is called by the bot with every message that is
     * sent to the IRC channel with the exception for messages
     * beginning with a command prefix, and messages sent by the bot
     * itself. If the implementing module does not care about the
     * communication in the channel, this method should immediately
     * return null without looking at the message.
     * @param message The message sent to the IRC channel.
     * @return A String containing a reply to the message, or null if
     * nothing should be sent to the IRC channel.
     * @see AbstractIRCBotModule
     * @see IRCMessage
     */
    public String parseMessage(IRCMessage message);

    /**
     * An IRC bot might think that it has been too quiet in a channel
     * for too long, or it may just feel like saying something at any
     * given time. To get something to say a random module is chosen
     * and its randomThought is invoked.
     * @param channel The channel that is too quiet.
     * @param sender The bot who asked for this thought.
     * @return A String containing a random thought or null if this
     * module has nothing nice to say.
     */
    public String randomThought(IRCChannel channel, Person sender);

    /**
     * getProbability is called to determine which module to choose to
     * express a random thought whenever the bot feels like doing
     * that.
     * @return An integer value between 0 and 100 which represents
     * the probability (percentage) that this module will be chosen to
     * express a random thought. If your module has a broad variety of
     * what is said and can be chosen several times in a row without
     * getting boring, you should return a high probability (100 or
     * just below). If your module does not have any random thoughts
     * return a low probability (0).
     */
    public int getProbability();

    /**
     * Each module can schedule events to occur at some point in the
     * future. When the future arrives and it is time for the event to
     * be executed, this method is called.
     * @param channel The channel in which the event is about to
     * occur.
     * @return A String containing the scheduled message to send to the
     * channel, or null if no message needs to be sent.
     */
    public String activateEvent(IRCChannel channel);

    /**
     * setModuleHandler is called by the module handler when a module
     * is loaded.
     * @param handler The module handler that is loading the module.
     */
    public void setModuleHandler(Modules handler);

    /**
     * loadState is called when a module is loaded into the bot. It
     * should read any settings or other external data needed for its
     * operation from the provided state. Note that the data to be
     * read has been saved by the saveState method in the same
     * class. This method is only supposed to read values that has
     * been saved by this class! If this module's saveState method
     * does not save any data, this method should be empty. Data is
     * read back from the state using the <code>SaveState.get</code>
     * method.
     * @param state The state to read from.
     * @see SaveState
     * @see AbstractIRCBotModule
     */
    public void loadState(SaveState state);

    /**
     * saveState is called when a module is unloaded from the bot or
     * whenever the IRC command !save is issued. Use it to save any
     * settings or other data needed for this modules operation.  If
     * this module does not keep any data that needs to be stored for
     * later sessions, this method should be empty.
     *
     * Use the <code>SaveState.add</code> method to put new data into
     * the state. Each data object that is to be stored in the state
     * should be tagged with a unique identifier. The identifier will
     * be padded with the name of the module so that identifiers in
     * different modules will not clash.
     * @param state The state to write to.
     * @see SaveState
     * @see AbstractIRCBotModule
     */
    public void saveState(SaveState state);

    /**
     * getDescription is called to get the information printed by the
     * !help <module name> command.
     * @param topic Either the string returned by
     * <code>getModuleName().toLowerCase()</code> or a member of
     * <code>getCommands()</code>. The convention says that the call
     * <code>getDescription(getModuleName().toLowerCase())</code>
     * should return a brief description about the module. This
     * information should not overlap with information given by calls
     * where the <code>topic</code> is a member of
     * <code>getCommands()</code>.
     * @param lang What language to use for the reply.
     * @return A String containing all the info needed to use this
     * module, tailored to the given topic. If the topic is
     * unexpected, the return value should be null.
     */
    public String getDescription(String topic, Language lang);

    /*
     * getAuthor is called by the !help <module name> command to get
     * the author of the module.
     * @return A String containing the name of the author of the
     * module.
     */
    public String getAuthor();
}
