package ircbot;

import java.io.*;
import java.util.*;
import ircbot.modules.*;

public class IRCChannel implements Serializable
{
    public static final long serialVersionUID = 4L;

    private String channelName;
    private String password;
    private long lastMess;
    private long minimumWait = Time.HOUR_3;
    private boolean online = false;
    private boolean verbose = true;
    private long silentTime = 0;
    private Language language = Language.SVENSKA;
    private boolean randomMess = false;
    private GreetingType greetingType = GreetingType.PERSONAL;
    private HashSet<Person> operators = new HashSet<Person>(8);
    private HashSet<Person> halfoperators = new HashSet<Person>(8);
    private HashSet<Person> voice = new HashSet<Person>(8);
    private String pendingText = null; // The remainder of the last message that was too long.
    private String lastReply = ""; // The last thing sent to this channel by the bot.
    private int repeatLimit = 2; // Do not send the same reply too many times
    private int repeatCount = 0;
    private ChannelStats stats = new ChannelStats();
    private HashSet<Person> users = new HashSet<Person>(32);

    public enum GreetingType { NONE, PERSONAL, ALWAYS }

    public IRCChannel(String name)
    {
        this(name, null);
    }

    public IRCChannel(String name, String password)
    {
        channelName = name.toLowerCase();
        this.password = password;
        lastMess = System.currentTimeMillis();
    }

    /*
    public void channelActive()
    {
        channelActive(minimumWait);
    }
    */

    //public void channelActive(long minWait)
    public void channelActive()
    {
        lastMess = System.currentTimeMillis();
        //minimumWait = minWait;
    }

    public String getName() { return channelName; }
    public String getPassword() { return password; }
    public long getLastMessage() { return lastMess; }
    public long getMinimumWait() { return minimumWait; }
    public Language getLanguage() { return language; }
    public boolean getRandomMessage() { return randomMess; }
    public GreetingType getGreetingType() { return greetingType; }
    public String getPendingText() { return pendingText; }
    public int getRepeatLimit() { return repeatLimit; }
    public int getRepeatCount() { return repeatCount; }
    public String getLastReply() { return lastReply; }
    public ChannelStats getStatistics() { return stats; }
    public Collection<Person> getOperators() { return operators; }
    public Collection<Person> getHalfop() { return halfoperators; }
    public Collection<Person> getVoice() { return voice; }
    public Collection<Person> getUsers() { return users; }

    public void setMinimumWait(long val) { minimumWait = val; }
    public void setLanguage(Language lang) { language = lang; }
    public void setRandomMessage(boolean val) { randomMess = val; }
    public void setGreetingType(GreetingType val) { greetingType = val; }
    public void setPendingText(String text) { pendingText = text; }
    public void setLastReply(String text) { lastReply = text; }
    public void setOnline(boolean val) { online = val; }

    public long setVerbose(boolean val)
    {
        long silent = 0;
        if ((!verbose) && val)
            silent = System.currentTimeMillis() - silentTime;
        else if (verbose && !val)
            silentTime = System.currentTimeMillis();

        verbose = val;
        return silent;
    }

    public boolean isVerbose() { return verbose; }
    public boolean isOnline() { return online; } // TODO: Join all onLine channels at startup.
    public boolean isOnline(Person p) { return users.contains(p); }
    public boolean isOperator(Person p) { return operators.contains(p); }
    public boolean isHalfop(Person p) { return halfoperators.contains(p); }
    public boolean hasVoice(Person p) { return voice.contains(p); }

    public boolean isInactive()
    {
        return System.currentTimeMillis() > getLastMessage() + getMinimumWait();
    }

    public String filterReply(String text)
    {
        if (text == null)
            return null;

        if (text.equals(getLastReply())) {
            repeatCount++;
            if (repeatCount >= repeatLimit) {
                text = "Känns som om jag sagt detta tidigare.. " + text;
            }
        }
        else {
            lastReply = text;
            repeatCount = 0;
        }

        return text;
    }

    public void addUser(Person p)
    {
        if (p != null && !users.contains(p))
            users.add(p);
    }

    public void removeUser(Person p)
    {
        users.remove(p);
    }

    public void addOperator(Person p)
    {
        if (!operators.contains(p))
            operators.add(p);
    }

    public void removeOperator(Person p)
    {
        operators.remove(p);
    }

    public void addHalfOperator(Person p)
    {
        if (!halfoperators.contains(p))
            halfoperators.add(p);
    }

    public void removeHalfOperator(Person p)
    {
        halfoperators.remove(p);
    }

    public void addVoice(Person p)
    {
        if (!voice.contains(p))
            voice.add(p);
    }

    public void removeVoice(Person p)
    {
        voice.remove(p);
    }

    public String toString()
    {
        return getName();
    }
}
