package ircbot;

import java.util.*;
import java.io.*;
import ircbot.modules.*;

public class Person implements Serializable, Comparable<Person>
{
    public static final long serialVersionUID = 6L;

    private String nick; /* Nick, case-sensitive */
    private ArrayList<String> aliases = new ArrayList<String>();
    private long score = 0;
    private Language lang = Language.SVENSKA;
    private String description = null;
    private String picture = null;
    private String greeting = null;

    private boolean op = false; /* Bot-op. Channel op is defined in IRCChannel. */
    private boolean trustee = false;
    private boolean ignore = false;

    private HashMap<String, PersonalStats> stats = new HashMap<String, PersonalStats>(4);

    /**
     * Create a new person.
     * @param nick
     */
    public Person(String nick)
    {
        this.nick = nick;
    }

    /**
     * Get the nick of this person.
     */
    public String getNick() { return nick; }
    
    /**
     * Get a collection of nicks used by this person. The default nick as returned by
     * getNick() is not included in this collection.
     */
    public Collection<String> getAliases() { return aliases; }

    /**
     * Get this persons score. A person can earn score by playing games or saying nice
     * things in the channels.
     */
    public long getScore() { return score; }

    /**
     * Get this persons preferred language.
     */
    public Language getLanguage() { return lang; }

    /**
     * Get a short description about this person. User editable.
     */
    public String getDescription() { return description; }

    /**
     * Get a url to an image showing this user. User editable.
     */
    public String getPicture() { return picture; }

    /**
     * Get the greeting phrase used to greet this person. User editable.
     */
    public String getGreeting() { return greeting; }

    /**
     * Get the number of words this person has sent in total. (All channels.)
     */
    public long getWords()
    {
        long words = 0;
        for (PersonalStats ps : stats.values()) {
            words += ps.getWords();
        }
        return words;
    }

    /**
     * Get the number of messages this person has sent in total. (All channels.)
     */
    public long getMessages()
    {
        long messages = 0;
        for (PersonalStats ps : stats.values()) {
            messages += ps.getMessages();
        }
        return messages;
    }

    public long getSmilies()
    {
        long smilies = 0;
        for (PersonalStats ps : stats.values()) {
            smilies += ps.getSmilies();
        }
        return smilies;
    }

    public long getSadies()
    {
        long sadies = 0;
        for (PersonalStats ps : stats.values()) {
            sadies += ps.getSadies();
        }
        return sadies;
    }

    public double getShoutFrequency()
    {
        double shout = 0;
        for (PersonalStats ps : stats.values()) {
            shout += ps.getShoutFrequency();
        }
        return shout / stats.values().size();
    }

    public Collection<Date> getWarnings()
    {
        ArrayList<Date> warnings = new ArrayList<Date>();
        for (PersonalStats ps : stats.values()) {
            warnings.addAll(ps.getWarnings());
        }
        return warnings;
    }

    public Date getFirstSeen()
    {
        Date first = null;
        for (PersonalStats ps : stats.values()) {
            if (first == null || first.getTime() > ps.getFirstSeen().getTime())
                first = ps.getFirstSeen();
        }
        return first;
    }

    public Date getLastActive()
    {
        Date last = null;
        for (PersonalStats ps : stats.values()) {
            if (last == null || last.getTime() < ps.getLastActive().getTime())
                last = ps.getLastActive();
        }
        return last;
    }

    public Date getLastLeave(IRCChannel channel)
    {
        PersonalStats ps = stats.get(channel.getName());
        if (ps == null)
            return null;
        return ps.getLastLeave();
    }

    public boolean isOperator() { return op; }
    public boolean isTrustee() { return trustee; }
    public boolean isIgnored() { return ignore; }

    public boolean isOnline( IRCChannel channel )
    {
        if (channel == null)
            return false;

        PersonalStats ps = stats.get(channel.getName());
        if (ps == null)
            return false;

        if ( ps.getLastLeave() == null ||
             ps.getLastLeave().getTime() < ps.getLastJoin().getTime() )
            return true;
        else
            return false;
    }

    public boolean isAuthorized(Modules.CommandType ct, IRCChannel channel)
    {
        switch (ct)
        {
            case PUBLIC: return true;
            case TRUSTEE: return op || trustee;
            case CHOP: return op || channel != null && channel.isOperator(this);
            case BOTOP: return op;
            default: return false;
        }
    }

    /**
     * Get this persons statistics for the given channel.
     * @param channel
     */
    public PersonalStats getStatistics( IRCChannel channel ) {
        return stats.get(channel.getName());
    }

    /**
     * Set the nick of this person. If the nick exists as an alias, it is removed from
     * the alias list. The old nick is thrown away.
     * @param newNick
     */
    public void setNick( String newNick )
    {
        removeAlias(newNick);
        nick = newNick;
    }

    public void setScore(int score) { this.score = score; }
    public void setLanguage(Language l) { lang = l; }
    public void setDescription(String desc) { description = desc; }
    public void setPicture(String url) { picture = url; }
    public void setGreeting(String text) { greeting = text; }
    public void setOperator(boolean val) { op = val; }
    public void setTrustee(boolean val) { trustee = val; }
    public void setIgnored(boolean val) { ignore = val; }

    public void incScore() { score++; }
    public void incScore(long s) { score += s; }
    public void decScore() { score--; }
    public void decScore(int s) { score -= s; }

    /**
     * Add an alias to this person. Will not add the alias if it is equal to this persons nick.
     * @param alias
     */
    public void addAlias(String alias)
    {
        if (!nick.equals(alias))
            aliases.add(alias);
    }

    public void removeAlias(String alias)
    {
        for (int i = 0; i < aliases.size(); i++)
            if (aliases.get(i).toLowerCase().equals(alias.toLowerCase())) {
                aliases.remove(i);
                return;
            }
    }

    public void mergeWith(Person p2)
    {
        // Nick and aliases
        for (String alias : p2.aliases)
            addAlias(alias);
        addAlias(p2.nick);

        // Score
        incScore(p2.score);

        // Description, picture etc.
        if (description == null)
            description = p2.description;
        if (picture == null)
            picture = p2.picture;
        if (greeting == null)
            greeting = p2.greeting;

        // Privileges
        op = op || p2.op;
        trustee = trustee || p2.trustee;

        // Statistics. Loop over all channels in p2, add or merge.
        Collection<String> channels = p2.stats.keySet();
        for (String channel : channels) {
            if (stats.containsKey(channel))
                stats.get(channel).mergeWith(p2.stats.get(channel));
            else
                stats.put(channel, p2.stats.get(channel));
        }
    }

    public void registerMessage(IRCMessage message)
    {
        PersonalStats ps = stats.get(message.getChannel().getName());
        String knownSmileys = ":) :-) ;) ;-) :] :-] :D :-D :} :-} lol \\o/";
        String knownSadies = ":( :-( :[ :-[ :{ :-{";
        int upper = 0;
        int chars = 0;

        ps.incMessages();
        ps.addWords(message.length());
        for (int i = 0; i < message.length(); i++) {
            String word = message.getParam(i);
            if (knownSmileys.indexOf(word) != -1)
                ps.incSmileys();
            else if (knownSadies.indexOf(word) != -1)
                ps.incSadies();

            chars += word.length();
            for (int c = 0; c < word.length(); c++)
                if (Character.isUpperCase(word.charAt(c)))
                    upper++;
        }
        ps.registerShouting(upper/(double)chars);

        if (Math.random() < 0.1)
            ps.setQuote(message.getMessage());
    }

    public int compareTo(Person p)
    {
        return (int)(p.getScore() - score);
    }

    public String toString()
    {
        return getNick();
    }

    public void notice(IRCChannel channel, People.NoticeType type)
    {
        if (channel == null)
            return;

        Date now = new Date();
        PersonalStats ps = stats.get(channel.getName());

        if (ps == null) {
            ps = new PersonalStats();
            stats.put(channel.getName(), ps);
        }

        switch (type) {
        case UNSEEN:
        case JOIN:
            ps.setLastJoin(now);
            break;

        case QUIT:
            ps.setLastLeave(now);
            break;

        case ACTION:
        case TYPE:
            ps.setLastActive(now);
            break;

        case GOOD:
            ps.setLastGood(now);
            incScore(Calc.rnd(1,5));
            break;

        case BAD:
            decScore(Calc.rnd(1,5));
            break;
        }
    }
}
