package ircbot.modules;

import ircbot.*;
import ircbot.modules.extras.*;
import java.io.*;
import java.util.*;

public class Calc extends AbstractIRCBotModule
{
    private HashMap<String, Double> variables = new HashMap<String, Double>();
    private ArrayList<Double> memory = new ArrayList<Double>();

    public Calc()
    {
        addCommand(Modules.CommandType.PUBLIC, "!calc");
        setupBuiltIn();
    }

    public String getModuleName()
    {
        return "Calculator";
    }

    public String action(IRCMessage message)
    {
        if (message.getBotCommand().equals("!calc") &&
            message.getParams() != null) {
            if (message.getParams().equals("reset")) {
                variables.clear();
                memory.clear();
                setupBuiltIn();
            }
            else
                return calc(message.getParams());
        }
        return null;
    }

    private void setupBuiltIn()
    {
        variables.put("pi", new Double(Math.PI));
        variables.put("e", new Double(Math.E));
    }

    private String calc(String something)
    {
        try {
            StringReader inputStream = new StringReader(filter(something));
            double a = readExpression(inputStream);

            if (inputStream.read() != -1)
                return "Syntax error";

            memory.add(0, new Double(a));
            if (memory.size() > 20)
                memory.remove(20);

            return "" + round(a);
        }
        catch (SyntaxErrorException se) {
            return "Syntax Error";
        }
        catch (UnknownIdentifyerException ide) {
            return "Unknown identifyer: " + ide.getID();
        }
        catch (IOException ioe) {
            return ioe.getMessage();
        }
    }

    public String getDescription(String topic, Language lang)
    {
        if (topic.equals(getModuleName().toLowerCase()))
            return Language.select(lang, "Beräknar matematiska uttryck. Beräknade svar kan sparas i variabler " +
                    "för senare åtkomst. Det går även att komma åt de senaste resultaten med $0, $1, $2 osv. " +
                    "Följande symboler och inbyggda funktioner stödjs: ",
                    "Evaluates mathematical expressions. Answers can be stored for later use in variables " +
                    "or can be accessed through $0, $1, $2, and so on. The " +
                    "following symbols and built in functions are supported: ") +
                "+ - / * ^ % ( ) = ~ & | < > && || << >> PI E sin cos tan asin acos atan sinh cosh tanh " +
                "log ln lg abs rnd sqrt cbrt";
        // x! == >= <= ++ -- += -= *= /=
        else if (topic.equals("!calc"))
            return "!calc " + Language.select(lang, "<uttryck> --- Beräknar matematiska uttryck.",
                                "<expression> --- Evaluates mathematical expressions.");
        else
            return null;
    }

    public String getAuthor()
    {
        return "Jesper Wilhelmsson";
    }

    private String filter(String something)
    {
        String result = "";

        for (int i = 0; i < something.length(); i++) {
            switch (something.charAt(i)) {
            case ' ':
            case '\t': break;
            default: result += Character.toLowerCase(something.charAt(i));
            }
        }
        return result;
    }

    private double readExpression(StringReader sr)
        throws IOException, CalcException
    {
        double a = readAdd(sr);
        int next;

        next = sr.read();
        while (next == '<' || next == '>' || next == '&' || next == '|') {
            sr.mark(1);
            if (next == '<') {
                next = sr.read();
                if (next == '<') {
                    sr.mark(1);
                    a = (double)((long)a << (long)readAdd(sr));
                }
                else {
                    sr.reset();
                    a = (a < readAdd(sr)) ? 1 : 0;
                }
            }
            else if (next == '>') {
                next = sr.read();
                if (next == '>') {
                    sr.mark(1);
                    a = (double)((long)a >> (long)readAdd(sr));
                }
                else {
                    sr.reset();
                    a = (a > readAdd(sr)) ? 1 : 0;
                }
            }
            else if (next == '&') {
                next = sr.read();
                if (next == '&') {
                    sr.mark(1);
                    a = ((readAdd(sr) != 0) && (a != 0)) ? 1 : 0;
                }
                else {
                    sr.reset();
                    a = (double)((long)a & (long)readAdd(sr));
                }
            }
            else if (next == '|') {
                next = sr.read();
                if (next == '|') {
                    sr.mark(1);
                    a = ((readAdd(sr) != 0) || (a != 0)) ? 1 : 0;
                }
                else {
                    sr.reset();
                    a = (double)((long)a | (long)readAdd(sr));
                }
            }
            next = sr.read();
        }
        sr.reset();
        return a;
    }

    private double readAdd(StringReader sr)
        throws IOException, CalcException
    {
        double a = readTerm(sr);
        int next;

        next = sr.read();
        while (next == '+' || next == '-') {
            sr.mark(1);
            if (next == '+')
                a += readTerm(sr);
            else
                a -= readTerm(sr);

            next = sr.read();
        }
        sr.reset();
        return a;
    }

    private double readTerm(StringReader sr)
        throws IOException, CalcException
    {
        double a = readPower(sr);
        int next;

        next = sr.read();
        while (next == '*' || next == '/') {
            sr.mark(1);
            if (next == '*')
                a *= readPower(sr);
            else
                a /= readPower(sr);

            next = sr.read();
        }
        sr.reset();
        return a;
    }

    private double readPower(StringReader sr)
        throws IOException, CalcException
    {
        double a = readFactor(sr);
        int next;

        next = sr.read();
        while (next == '^' || next == '%') {
            sr.mark(1);
            if (next == '^')
                a = Math.pow(a,readFactor(sr));
            else
                a = a % readFactor(sr);

            next = sr.read();
        }
        sr.reset();
        return a;
    }

    private double readFactor(StringReader sr)
        throws IOException, CalcException
    {
        int next = sr.read();

        /* ( expr ) */
        if (next == '(') {
            sr.mark(1);
            double a = readExpression(sr);
            next = sr.read();
            if (next != ')') {
                throw new SyntaxErrorException();
            }
            sr.mark(1);
            return a;
        }

        /* - factor */
        if (next == '-') {
            sr.mark(1);
            return -readFactor(sr);
        }

        /* ~ factor */
        if (next == '~') {
            sr.mark(1);
            return ~(long)readFactor(sr);
        }

        /* ! factor */
        if (next == '!') {
            sr.mark(1);
            return (((long)readFactor(sr)) == 0) ? 1 : 0;
        }

        /* $n */
        if (next == '$') {
            sr.mark(1);
            int a = (int)readNumber(sr);
            if (a >= memory.size())
                throw new UnknownIdentifyerException("$" + a);
            return memory.get(a);
        }

        /* number */
        if (Character.isDigit(next)) {
            sr.reset();
            return readNumber(sr);
        }

        /* identifier */
        if (Character.isLetter(next)) {
            String ident;
            sr.reset();
            ident = readIdentifyer(sr);

            if (ident.equals("sin")) return Math.sin(readFactor(sr));
            if (ident.equals("cos")) return Math.cos(readFactor(sr));
            if (ident.equals("tan")) return Math.tan(readFactor(sr));
            if (ident.equals("sinh")) return Math.sinh(readFactor(sr));
            if (ident.equals("cosh")) return Math.cosh(readFactor(sr));
            if (ident.equals("tanh")) return Math.tanh(readFactor(sr));
            if (ident.equals("asin")) return Math.asin(readFactor(sr));
            if (ident.equals("acos")) return Math.acos(readFactor(sr));
            if (ident.equals("atan")) return Math.atan(readFactor(sr));

            if (ident.equals("log")) return Math.log10(readFactor(sr));
            if (ident.equals("lg")) return Math.log(readFactor(sr)) / Math.log(2);
            if (ident.equals("ln")) return Math.log(readFactor(sr));

            if (ident.equals("abs")) return Math.abs(readFactor(sr));
            if (ident.equals("sqrt")) return Math.sqrt(readFactor(sr));
            if (ident.equals("cbrt")) return Math.cbrt(readFactor(sr));
            if (ident.equals("rnd")) return rnd((long)readFactor(sr));

            next = sr.read();
            if (next == '=') {
                sr.mark(1);
                variables.put(ident,new Double(readExpression(sr)));
            }
            else
                sr.reset();

            if (variables.get(ident) != null) {
                return variables.get(ident).doubleValue();
            }

            throw new UnknownIdentifyerException(ident);
        }

        throw new SyntaxErrorException();
    }


    private double readNumber(StringReader sr) throws IOException
    {
        double a = 0;
        int next = sr.read();

        while (Character.isDigit(next)) {
            sr.mark(1);
            a = a * 10 + (next - '0');
            next = sr.read();
        }

        if (next == '.' || next == ',') {
            int frac = 10;
            sr.mark(1);
            next = sr.read();
            while(Character.isDigit(next)) {
                sr.mark(1);
                a += (next - '0') / (float)frac;
                frac *= 10;
                next = sr.read();
            }
        }

        sr.reset();
        return a;
    }

    private String readIdentifyer(StringReader sr) throws IOException
    {
        String result = "";
        int next = sr.read();

        while (Character.isLetter(next) ||
               Character.isDigit(next)) {
            sr.mark(1);
            result += (char)next;
            next = sr.read();
        }

        sr.reset();
        return result;
    }


    public static double round(double n)
    {
        return ((long)(n * 1000000 + 0.5)) / 1000000.0;
    }

    /**
     * @return true if the given string contains only digits, false
     * otherwise.
     */
    public static boolean isNumber(String s)
    {
        for (int i = 0; i < s.length(); i++)
            if (!Character.isDigit(s.charAt(i)))
                return false;

        return true;
    }

    /**
     * @return The first positive integer found in the string or -1 if no
     * numbers were found.
     */
    public static int extractInt(String s)
    {
        if (s == null)
            return -1;

        int start = 0;
        int l = s.length();

        while (start < l && !Character.isDigit(s.charAt(start)))
            start++;

        if (start == l)
            return -1;

        int end = start + 1;
        while (end < l && Character.isDigit(s.charAt(end)))
            end++;

        return Integer.parseInt(s.substring(start, end));
    }

    /**
     * @param max The upper bound of the random number.
     * @return A random number between (including) zero and (not
     * including) <code>max</code>.
     */
    public static int rnd(int max)
    {
        return (int)rnd(0, (long)max - 1);
    }

    /**
     * @param max The upper bound of the random number.
     * @return A random number between (including) zero and (not
     * including) <code>max</code>.
     */
    public static long rnd(long max)
    {
        return rnd(0, max - 1);
    }

    /**
     * @param min The lower bound of the random number.
     * @param max The upper bound of the random number.
     * @return A random number between (including) <code>min</code>
     * and (including) <code>max</code>.
     */
    public static int rnd(int min, int max)
    {
        return (int)rnd((long)min, (long)max);
    }

    /**
     * @param min The lower bound of the random number.
     * @param max The upper bound of the random number.
     * @return A random number between (including) <code>min</code>
     * and (including) <code>max</code>.
     */
    public static long rnd(long min, long max)
    {
        return (long)(Math.random() * ((max + 1) - min)) + min;
    }
}
