package ircbot.modules;

import java.io.*;
import java.util.*;
import javax.swing.filechooser.*;
import ircbot.*;

/**
 * Module handler. This module is responsible for all communication
 * between the IRC engine and other modules.  The IRC engine will only
 * see one module, preferably this one. The module distributes
 * commands, IRC messages and events to all modules (including it
 * self). It also handles loading and unloading of modules. Unloading
 * this module is a bad idea.
 */
public class Modules extends AbstractIRCBotModule
{
    public static enum CommandType { PUBLIC, TRUSTEE, CHOP, BOTOP }

    private ArrayList<IRCBotModule> modules = new ArrayList<IRCBotModule>();
    private IRCBot myBot;
    private SaveState state;
    private boolean recursive;
    private long lastSave;

    public Modules( IRCBot bot )
    {
        myBot = bot;

        // TODO: !mute to mute a single module. Command added to all modules automatically.
        addCommand( CommandType.PUBLIC, "!modules" );
        addCommand( CommandType.PUBLIC, "!commands" );
        addCommand( CommandType.PUBLIC, "!more" );

        addCommand( CommandType.BOTOP, "!load" );
        addCommand( CommandType.BOTOP, "!unload" );
        addCommand( CommandType.BOTOP, "!reload" );
        addCommand( CommandType.BOTOP, "!save" );

        state = new SaveState(myBot.getName());
        recursive = false;
        lastSave = System.currentTimeMillis();

        addModule(this);
    }

    public IRCBot getBot()
    {
        return myBot;
    }

    public String getModuleName()
    {
        return "Modules";
    }

    public String action(IRCMessage message)
    {
        if (recursive) {
            String params = message.getParams();

            if (message.getBotCommand().equals("!modules")) {
                LinkedList<String> names = new LinkedList<String>();
                for (IRCBotModule mod : getModules())
                    names.add(mod.getModuleName());
                return "Laddade moduler: " + IRCBot.makeList(names);
            }

            else if (message.getBotCommand().equals("!commands")) {
                LinkedList<String> names = new LinkedList<String>();
                for (IRCBotModule mod : getModules())
                    names.addAll(mod.getCommands(CommandType.PUBLIC));
                if (message.getSender().isOperator() || message.getChannel().isOperator(message.getSender()))
                    for (IRCBotModule mod : getModules())
                        names.addAll(mod.getCommands(CommandType.CHOP));
                if (message.getSender().isOperator() || message.getSender().isTrustee())
                    for (IRCBotModule mod : getModules())
                        names.addAll(mod.getCommands(CommandType.TRUSTEE));
                if (message.getSender().isOperator())
                    for (IRCBotModule mod : getModules())
                        names.addAll(mod.getCommands(CommandType.BOTOP));

                return "Tillgängliga kommandon: " + IRCBot.makeList(names);
            }

            else if (message.getBotCommand().equals("!more")) {
                IRCChannel ch = message.getChannel();
                if (ch != null) {
                    String pt = ch.getPendingText();
                    if (pt != null) {
                        if (pt.length() <= 900)
                            ch.setPendingText(null);
                        return pt;
                    }
                }
                else
                    return "!more is currently only working within a channel.";
            }

            else if (message.getBotCommand().equals("!load")) {
                loadModule(params);
            }

            else if (message.getBotCommand().equals("!unload")) {
                unloadModule(params);
            }

            else if (message.getBotCommand().equals("!reload")) {
                String result = unloadModule(params);
                if (params != null)
                    result += " -> " + loadModule(params);
                return result;
            }

            else if (message.getBotCommand().equals("!save")) {
                return saveState();
            }

            return null;
        }
        else {
            ArrayList<String> ambiguous = new ArrayList<String>();
            IRCBotModule module = null;

            for (IRCBotModule mod : getModules()) {
                for (CommandType ct : CommandType.values()) {
                    if (message.getSender().isAuthorized(ct, message.getChannel())) {
                        for (String cmnd : mod.getCommands(ct)) {
                            if (cmnd.startsWith(message.getBotCommand())) {
                                ambiguous.add(cmnd.charAt(0) + mod.getModuleName() + "." + cmnd.substring(1));
                                module = mod;
                            }
                            else if ( message.getBotCommand().startsWith(cmnd.charAt(0) +
                                      mod.getModuleName().toLowerCase() + ".") ) {
                                String suffix = message.getBotCommand().substring(message.getBotCommand().indexOf('.') + 1);
                                if (cmnd.substring(1).startsWith(suffix)) {
                                    ambiguous.add(cmnd.charAt(0) + mod.getModuleName() + "." + cmnd.substring(1));
                                    module = mod;
                                }
                            }
                        }
                    }
                }
            }

            if ( ambiguous.size() == 1 ) {
                try {
                    recursive = true;
                    String command = ambiguous.get(0);
                    message.replaceBotCommand(command.charAt(0) + command.substring(command.indexOf('.') + 1));
                    String result = module.action( message );
                    recursive = false;
                    if (result != null && result.length() > 900) {
                        int space = result.lastIndexOf(' ', 880);
                        IRCChannel ch = message.getChannel();
                        if (ch != null)
                            ch.setPendingText(result.substring(space));
                        result = result.substring(0, space) + "... [!more to read]";
                    }
                    return result;
                }
                catch (Exception e) {
                    recursive = false;
                    return "Error (command) in module " + module.getModuleName() + ": " + e.getMessage();
                }
            }
            else if (ambiguous.size() > 1) {
                return "Ambiguous command: " + IRCBot.makeList(ambiguous);
            }
        }

        return null;
    }

    public String parseMessage(IRCMessage message)
    {
        if (!recursive) {
            String reply = "";
            recursive = true;
            for (IRCBotModule mod : getModules()) {
                String txt;
                try {
                    txt = mod.parseMessage(message);
                }
                catch (Exception e) {
                    txt = "Error (parse) in module " + mod.getModuleName() +
                        ": " + e.getMessage();
                }
                if (txt != null) {
                    /* Uncomment to add [modulename] in front of replies. */
                    //if (!txt.startsWith("\001ACTION") &&
                    //    !txt.startsWith("syscmd:"))
                    //    reply += "[" + mod.getModuleName() + "] ";
                    reply += txt + "\n";
                }
            }
            recursive = false;
            if (reply.length() > 0)
                return reply;
        }

        return null;
    }

    public String randomThought(IRCChannel channel, Person sender)
    {
        String reply = null;

        if (!recursive) {
            recursive = true;

            LinkedList<IRCBotModule> list =
                new LinkedList<IRCBotModule>(getModules());
            Collections.sort(list);
            for (IRCBotModule mod : list) {
                try {
                    reply = mod.randomThought(channel, sender);
                }
                catch (Exception e) {
                    reply = "Error (random) in module " + mod.getModuleName() +
                        ": " + e.getMessage();
                }
                if (reply != null)
                    break;
            }
            recursive = false;
        }
        return reply;
    }


    /**********************************************************************
     * Module handling
     */

    public void addModule(IRCBotModule mod)
    {
        modules.add(mod);
        mod.setModuleHandler(this);
        loadState(mod);
    }

    public IRCBotModule getModule(String name)
    {
        for (IRCBotModule mod : modules) {
            if (mod.getModuleName().toLowerCase().equals(name.toLowerCase()))
                return mod;
        }
        return null;
    }

    public Collection<IRCBotModule> getModules()
    {
        return modules;
    }

    private String loadModule(String params)
    {
        if (params == null) {
            try {
                FileSystemView fs = FileSystemView.getFileSystemView();
                File f = fs.createFileObject(fs.getDefaultDirectory(),
                                             "/bottenanna/ircbot/modules");
                File[] files = fs.getFiles(f, false);
                LinkedList<String> names = new LinkedList<String>();

                for (File mod : files) {
                    String name = mod.getName();
                    if (name.endsWith(".class"))
                        names.add(name.substring(0, name.indexOf(".class")));
                }
                return "Tillgängliga moduler: " +
                    IRCBot.makeList(names);
            }
            catch (Exception e) {
                String error = "[!load: " + e.getClass() + "] " + e.getMessage();
                System.err.println(error);
                return error;
            }
        }
        else {
            if (params.toLowerCase().equals("modules") ||
                params.toLowerCase().equals("channels") ||
                params.toLowerCase().equals("people"))
                return "Det går inte att ladda in " + params;

            try {
                Class<?> c = Class.forName("ircbot.modules." + params);
                IRCBotModule mod = (IRCBotModule)c.newInstance();
                addModule(mod);
                return mod.getModuleName() + " laddad.";
            }
            catch (Exception e) {
                System.err.println("[!load: " + e.getClass() + "] " +
                                   e.getMessage());
            }
            return "Hittar inte modulen '" + params + "'";
        }
    }

    private String unloadModule(String params)
    {
        if (params == null) {
            HashSet<String> names = new HashSet<String>();
            for (IRCBotModule mod : getModules())
                names.add(mod.getModuleName());
            return "Laddade moduler: " + IRCBot.makeList(names);
        }
        else {
            if (params.toLowerCase().equals("modules") ||
                params.toLowerCase().equals("channels") ||
                params.toLowerCase().equals("people"))
                return "Det går inte att ta bort " + params;

            Iterator<IRCBotModule> it = getModules().iterator();
            while (it.hasNext()) {
                IRCBotModule mod = it.next();
                if (mod.getModuleName().toLowerCase().
                    equals(params.toLowerCase())) {
                    saveState(mod);
                    it.remove();
                    return params + " borttagen.";
                }
            }
            return "Hittar inte modulen '" + params + "'";
        }
    }

    /**********************************************************************
     * Saving and loading preferences and other stuff.
     */

    public String periodicSave()
    {
        if (System.currentTimeMillis() > lastSave + Time.HOUR_24)
            return saveState();
        else
            return null;
    }

    public String saveState()
    {
        for (IRCBotModule mod : modules) {
            state.setModule(mod.getModuleName());
            mod.saveState(state);
        }

        state.save(myBot.getName() + ".state");
        lastSave = System.currentTimeMillis();
        return "State saved " + Time.date2str(new Date());
    }

    private void saveState(IRCBotModule mod)
    {
        state.setModule(mod.getModuleName());
        mod.saveState(state);
        state.save(myBot.getName() + ".state");
    }

/*
    private void loadState()
    {
        for (IRCBotModule mod : modules) {
            state.setModule(mod.getModuleName());
            mod.loadState(state);
        }
    }
*/

    private void loadState(IRCBotModule mod)
    {
        state.setModule(mod.getModuleName());
        mod.loadState(state);
    }

    public String getDescription(String topic, Language lang)
    {
        if (topic.equals(getModuleName().toLowerCase())) {
            TreeSet<String> moduleNames = new TreeSet<String>();
            for (IRCBotModule mod : modules)
                moduleNames.add(mod.getModuleName());
            return "Modulhanterare. Denna modul ansvarar för all " +
                "kommunikation mellan IRC-motorn och modulerna. Den " +
                "distribuerar kommandon, IRC-meddelanden och händelser i " +
                "kanalen till samtliga moduler (inklusive sig själv). Den " +
                "är också ansvarig för dynamisk laddning och borttagning " +
                "av moduler. " +
                "Moduler laddade just nu: " + moduleNames;
        }
        else if (topic.equals("!modules"))
            return "!modules --- Visa en lista över alla laddade moduler.";
        else if (topic.equals("!commands"))
            return "!commands --- Visa en lista över alla tillgängliga " +
                "kommandon.";
        else if (topic.equals("!more"))
            return "!more --- Om ett kommando resulterar i ett svar som " +
                "överstiger 900 tecken kommer det att klippas av. !more " +
                "används då för att visa nästa del av svaret.";
        else if (topic.equals("!load"))
            return "!load [module] --- Ladda in en ny modul. Om ingen " +
                "modul anges listas samtliga tillgängliga moduler.";
        else if (topic.equals("!unload"))
            return "!unload [module] --- Ta bort angiven modul. Om ingen " +
                "modul anges listas alla laddade moduler.";
        else if (topic.equals("!reload"))
            return "!reload [module] --- Ladda om angiven modul. Om ingen " +
                "modul anges listas alla laddade moduler. Tänk på att en " +
                "en modul inte nödvändigtvis har samma namn som dess " +
                "klassfil. Om dessa inte är identiska kan !reload inte " +
                "användas. Använd då !unload/!load.";
        else if (topic.equals("!save"))
            return "!save [module] --- Spara inställningar och data för " +
                "angiven modul. Om ingen modul anges sparas alla moduler.";
        else
            return null;
    }

    public String getAuthor()
    {
        return "Jesper Wilhelmsson";
    }
}
