package ircbot.modules;

import java.util.*;
import ircbot.*;

public class People extends AbstractIRCBotModule
{
    public enum NoticeType { JOIN, QUIT, ACTION, TYPE, GOOD, BAD, UNSEEN }

    private HashMap<String, Person> people;
    private HashSet<String> removed;

    public People()
    {
        addCommand(Modules.CommandType.PUBLIC, "!seen");
        addCommand(Modules.CommandType.PUBLIC, "!info");
        addCommand(Modules.CommandType.PUBLIC, "!score");
        addCommand(Modules.CommandType.PUBLIC, "!highscore");
        addCommand(Modules.CommandType.PUBLIC, "!rename");
        addCommand(Modules.CommandType.PUBLIC, "!description");
        addCommand(Modules.CommandType.PUBLIC, "!greeting");
        addCommand(Modules.CommandType.PUBLIC, "!alias");

        addCommand(Modules.CommandType.TRUSTEE, "!reward");
        addCommand(Modules.CommandType.TRUSTEE, "!punish");

        addCommand(Modules.CommandType.BOTOP, "!botop");
        addCommand(Modules.CommandType.BOTOP, "!trustee");
        addCommand(Modules.CommandType.BOTOP, "!ignore");
        addCommand(Modules.CommandType.BOTOP, "!remove");
    }

    public String getModuleName()
    {
        return "People";
    }

    public String action( IRCMessage message )
    {
        String command = message.getBotCommand();

        if (command.equals("!seen")) {
            // TODO: !seen on someone ignored should reply "I really don't care where n is!"
            int firstNick = 0;
            IRCChannel channel = ((Channels)(getModuleHandler().getModule("channels"))).getChannel(message.getParam(0));
            if (channel != null) {
                if ( !channel.isOnline(message.getSender()) && !message.getSender().isOperator() )
                    return "You are not present in that channel";
                firstNick++;
            } else if ( message.length() > 0 && message.getParam(0).toLowerCase().equals("all") ) {
                if (!message.getSender().isOperator())
                    return "You are not privileged enough to do that.";
                firstNick++;
            } else {
                channel = message.getChannel();
                if (channel == null)
                    return "Please specify an existing channel";
            }

            /* if channel == null below, 'all' was specified by
             * authorized personnel.
             */

            if ( message.getParam(firstNick) != null ) {
                /* One or more nicks has been specified. */
                String result = "";
                while ( message.getParam(firstNick) != null ) {
                    Person p = getPerson(message.getParam(firstNick));
                    if ( p != null ) {
                        if ( p.isOnline(channel) )
                            result += p.getNick() + " är online. ";
                        Date lastActive;
                        if (channel != null)
                            lastActive = p.getStatistics(channel).getLastActive();
                        else
                            lastActive = p.getLastActive();
                        result += Language.select(message.getSender().getLanguage(),
                                                  "Senast jag såg något livstecken från " + p.getNick() + " var ",
                                                  "Last time I saw any sign of life from " + p.getNick() + " was ") +
                            Time.date2str(lastActive) + '\n';
                    } else {
                        result += Language.select(message.getSender().getLanguage(), "\001ACTION har aldrig sett '",
                                                  "\001ACTION has never seen '") + message.getParam(firstNick) + "'\001\n";
                    }
                    firstNick++;
                }
                return result;
            } else {
                /* No nick was specified, show all seen users in the
                 * appropriate channel(s).
                 */
                TreeSet<String> names = new TreeSet<String>();
                for (Person pers : people.values()) {
                    if (channel == null || pers.isOnline(channel))
                        names.add(pers.getNick());
                }
                return Language.select(message.getSender().getLanguage(), "Jag har sett: ", "I have seen: ") +
                    IRCBot.makeList(names);
            }
        }

        else if (command.equals("!info")) {
            Person p;
            if (message.length() == 0)
                p = message.getSender();
            else
                p = getPerson(message.getParam(0));

            if (p != null) {
                String reply = p.getNick();
                if (p.getAliases().size() > 0)
                    reply += " a.k.a. " + IRCBot.makeList(p.getAliases()) + ",";

                if (p.getDescription() == null)
                    reply += Language.select(message.getSender().getLanguage(),
                            " är en person som inte har provat kommandot !description.",
                            " is a person who has not tried the !description command.");
                else
                    reply += " " + p.getDescription() + ".";

                reply += " " + p.getNick() + Language.select(message.getSender().getLanguage(),
                        " har skrivit " + p.getMessages() + " meddelanden med totalt " +
                        p.getWords() + " ord sedan ",
                        " has written " + p.getMessages() + " messages using " + p.getWords() +
                        " words in total since ") +
                    Time.date2str(p.getFirstSeen()) + ".";

                if (message.getChannel() != null) {
                    PersonalStats ps = p.getStatistics(message.getChannel());
                    String type = Time.date2str(ps.getLastActive());
                    if (type != null)
                        reply += Language.select(message.getSender().getLanguage(),
                                             " Senast " + p.getNick() + " var aktiv var " + type + ".",
                                             " Last time " + p.getNick() + " was active was " + type + ".");
                    String good = Time.date2str(ps.getLastGood());
                    if (good != null)
                        reply += Language.select(message.getSender().getLanguage(),
                                                 " Senast " + p.getNick() + " sa något intressant var " + good + ".",
                                                 " Last time " + p.getNick() + " said anything interesting was " +
                                                 good + ".");
                    reply += " Max idle: " +
                    Time.time2str(ps.getMaxIdle());
                }

                return reply;
            }
            else if (message.length() > 0) {
                return "\001ACTION " + Language.select(message.getSender().getLanguage(),
                        "har aldrig sett '", "has never seen '") + message.getParam(0) + "'\001";
            }
            return null;
        }

        else if (command.equals("!score")) {
            Person p = message.getSender();
            return p.getNick() +
                Language.select(message.getSender().getLanguage(), " har " + p.getScore() + " poäng",
                                "'s score is " + p.getScore());
        }

        else if (command.equals("!highscore")) {
            ArrayList<Person> folks = new ArrayList<Person>(people.values());
            ArrayList<String> l = new ArrayList<String>();
            Person last = null;

            Collections.sort(folks);
            for (int i = 0; i < folks.size() && l.size() < 5; i++) {
                Person p = folks.get(i);
                if (p != last) {
                    last = p;
                    l.add(p.getNick() + ": " + p.getScore());
                }
            }
            return IRCBot.makeList(l, false);
        }

        else if (command.equals("!rename") && message.length() > 0) {
            Person p = message.getSender();
            String oldNick = p.getNick();
            p.setNick(message.getParam(0));
            p.addAlias(oldNick);
        }

        else if (command.equals("!description") && message.length() > 0) {
            message.getSender().setDescription(message.getParams());
        }

        else if (command.equals("!botop")) {
            if (message.length() == 0) {
                LinkedList<String> operators = new LinkedList<String>();
                for (Person p : people.values())
                    if (p.isOperator())
                        operators.add(p.getNick());
                return Language.select(message.getSender().getLanguage(),
                        "Följande personer tar jag order från: ", "These are my masters: ") +
                        IRCBot.makeList(operators);
            } else if ( message.getParam(0).toLowerCase().equals("remove") ) {
                if (message.length() == 1) {
                    return Language.select(message.getSender().getLanguage(), 
                            "Tala gärna om vem det är du vill degradera.",
                            "Don't hessitate to tell me who it is that you want to degrade.");
                } else if ( getPerson(message.getParam(1)).isOperator() ) {
                    if (message.getParam(1).equals("joppe")) {
                        return Language.select(message.getSender().getLanguage(),
                                "Nej, det vill jag inte.", "Yeah, right.");
                    }
                    getPerson(message.getParam(1)).setOperator(false);
                    return message.getParam(1) + Language.select(message.getSender().getLanguage(),
                            " är inte längre botop", " has been degraded");
                } else {
                    return message.getParam(1) + Language.select(message.getSender().getLanguage(),
                            " är inte botop", " is a NOP");
                }
            } else {
                Person p = getPerson(message.getParam(0));
                if (p != null && p.isOperator())
                    return message.getParam(0) + Language.select(message.getSender().getLanguage(),
                            " är redan botop", " is already bot operator");
                else if (p != null) {
                    p.setOperator(true);
                    return message.getParam(0) + Language.select(message.getSender().getLanguage(),
                            " befodras nu till botop", " is promoted to bot operator");
                }
            }
        }

        else if (command.equals("!trustee")) {
            if (message.length() == 0) {
                LinkedList<String> trustees = new LinkedList<String>();
                for (Person p : people.values())
                    if (p.isTrustee())
                        trustees.add(p.getNick());
                return Language.select(message.getSender().getLanguage(),
                        "Följande personer är pålitliga: ", "These are my trustees: ") +
                        IRCBot.makeList(trustees);
            } else if ( message.getParam(0).toLowerCase().equals("remove") ) {
                if (message.length() == 1) {
                    return Language.select(message.getSender().getLanguage(), 
                            "Tala gärna om vem det är du vill degradera.",
                            "Don't hessitate to tell me who it is that you want to degrade.");
                } else if ( getPerson(message.getParam(1)).isTrustee() ) {
                    if (message.getParam(1).equals("joppe")) {
                        return Language.select(message.getSender().getLanguage(),
                                "Nej, det vill jag inte.", "Yeah, right.");
                    }
                    getPerson(message.getParam(1)).setTrustee(false);
                    return message.getParam(1) + Language.select(message.getSender().getLanguage(),
                            " är inte längre pålitlig", " has been degraded");
                } else {
                    return message.getParam(1) + Language.select(message.getSender().getLanguage(),
                            " är inte pålitlig", " aint a trustee");
                }
            } else {
                Person p = getPerson(message.getParam(0));
                if (p != null && p.isTrustee())
                    return message.getParam(0) + Language.select(message.getSender().getLanguage(),
                            " är redan pålitlig", " is already a trustee");
                else if (p != null) {
                    p.setTrustee(true);
                    return message.getParam(0) + Language.select(message.getSender().getLanguage(),
                            " är nu pålitlig", " is promoted to trustee");
                }
            }
        }

        else if (command.equals("!reward") && message.length() > 0) {
            String nick = message.getParam(0);
            int points;

            if (message.length() > 1) {
                points = Calc.extractInt(message.getParam(1));
            } else {
                points = Calc.rnd(20, 50);
            }

            Person p = getPerson(nick);
            if (p != null && points != -1)
                p.incScore(points);
        }

        else if (command.equals("!punish") && message.length() > 0) {
            String nick = message.getParam(0);
            int points;

            if (message.length() > 1) {
                points = Calc.extractInt(message.getParam(1));
            } else {
                points = Calc.rnd(20, 50);
            }

            Person p = getPerson(nick);
            if (p != null && points != -1)
                p.decScore(points);
        }

        else if (command.equals("!alias")) {
            String nick;
            if (message.length() == 0)
                nick = message.getSender().getNick();
            else
                nick = message.getParam(0);

            Person p1 = getPerson(nick.toLowerCase());
            if (p1 == null)
                return Language.select(message.getSender().getLanguage(), "Jag känner inte ", "I don't know ") + nick;

            if (message.length() > 1) {
                String alias = message.getParam(1);
                Person p2 = getPerson(alias.toLowerCase());

                if (p2 == null) {
                    p1.addAlias(alias);
                    people.put(alias.toLowerCase(), p1);
                } else if (p1 != p2) {
                    p1.mergeWith(p2);
                    people.put(alias.toLowerCase(), p1);
                } else if (p1.getNick().toLowerCase().equals(alias.toLowerCase())) {
                    p1.setNick(nick);
                    p1.addAlias(alias);
                }
            } else
                return getPerson(nick) + " a.k.a. " + IRCBot.makeList(getPerson(nick).getAliases());
        }

        else if (command.equals("!ignore")) {
            boolean remove = false;
            int firstNick = 0;
            if ( message.length() > 1 && message.getParam(0).equals("remove") ) {
                remove = true;
                firstNick++;
            }
            for (int i = firstNick; i < message.length(); i++) {
                Person p = getPerson(message.getParam(i));
                if (p != null) {
                    p.setIgnored(!remove);
                    if (!remove) {
                        return p.getNick() + Language.select(message.getSender().getLanguage(),
                                " finns även i databasen. Använd !remove för att ta bort personen helt.",
                                " is also present in the data base. Use !remove to erase all occurances.");
                    }
                }
            }
            ArrayList<Person> ignores = new ArrayList<Person>();
            for (Person p : people.values()) {
                if (p.isIgnored())
                    ignores.add(p);
            }
            return "Ignores: " + IRCBot.makeList(ignores);
        }

        else if (command.equals("!remove")) {
            if (message.length() > 0) {
                String nick = message.getParam(0);
                Person p = getPerson(nick);
                if (p != null) {
                    if (!p.getNick().toLowerCase().equals(nick.toLowerCase())) {
                        p.removeAlias(nick);
                    }
                    else if (p.getAliases().size() > 0) {
                        Iterator<String> it = p.getAliases().iterator();
                        p.setNick(it.next());
                    }
                    people.remove(nick.toLowerCase());
                    removed.add(nick.toLowerCase());
                    return nick + Language.select(message.getSender().getLanguage(),
                                        " är nu bortplockad ur databasen.",
                                        " has been erased from my memory.");
                }
            }
        }

        return null;
    }

    public String parseMessage(IRCMessage message)
    {
        if (message.getCommand().equals("privmsg")) {
            String mess = message.getMessage().toLowerCase();
            message.getSender().notice(message.getChannel(), NoticeType.TYPE);
            if (message.getChannel() != null)
                message.getSender().registerMessage(message);

            // TODO: 'vem är ...?'  Stop after '?' or at most two words.
            if (mess.contains(" vem är ") || mess.startsWith("vem är ")) {
                int end = mess.length();
                if (mess.endsWith("?"))
                    end--;
                String nick = mess.substring(mess.indexOf("vem är ") + 7, end);
                if (nick.equals("jag"))
                    nick = message.getSender().getNick();
                return action(new IRCMessage(message, "!info", nick));
            }

            if (mess.contains(" var är ") || mess.startsWith("var är ")) {
                int end = mess.length();
                if (mess.endsWith("?"))
                    end--;
                String nick = mess.substring(mess.indexOf("var är ") + 7, end);
                if (nick.equals("jag"))
                    return "Du är i kanalen " + message.getChannel().getName();
                return action(new IRCMessage(message, "!seen", nick));
            }
        }

        else if (message.getCommand().equals("join")) {
            IRCChannel channel = message.getChannel();
            message.getSender().notice(channel, NoticeType.JOIN);

            if ( channel != null &&
                 (message.getSender().getLastLeave(channel) == null ||
                  System.currentTimeMillis() - Time.MIN_30 > message.getSender().getLastLeave(channel).getTime()) &&
                 channel.getGreetingType() != IRCChannel.GreetingType.NONE ) {
                String txt = "";
                // TODO: Remember user name from host and use it to find Person-objects. (15:46:06) Safir_anon856 [~username@some.server.net] kom in i rummet.
                if ( message.getSender().isOperator() )
                    txt = "Master " + message.getSender() + " returns to the dungeon.\n";
                if ( message.getSender().getGreeting() != null )
                    return txt + message.getSender().getGreeting();
                else if ( channel.getGreetingType() == IRCChannel.GreetingType.ALWAYS && txt.equals("") ) {
                    txt = Time.timeComment(new Date(), message.getSender().getNick());
                }
            }
        }

        else if (message.getCommand().equals("quit")) {
            message.getSender().notice(message.getChannel(), NoticeType.QUIT);
        }

        else if (message.getCommand().equals("nick")) {
            message.getSender().notice(message.getChannel(), NoticeType.ACTION);
            action(new IRCMessage(message, "!alias", message.getSender().getNick() + " " + message.getMessage()));
        }

        else if (message.getCommand().equals("353")) {
            /* People already in the channel at arrival. */
            String[] names = message.getMessage().split(" ");
            for (int i = 0; i < names.length; i++) {
                String nick = names[i];
                if ((names[i].charAt(0) == '@') || (names[i].charAt(0) == '%') || (names[i].charAt(0) == '+'))
                    nick = names[i].substring(1);
                notice(nick, message.getChannel(), NoticeType.UNSEEN);
            }
        }

        return null;
    }

    public Person notice(String nick, IRCChannel channel, NoticeType type)
    {
        Person p = people.get(nick.toLowerCase());

        if (p == null) {
            p = new Person(nick);
            if (!removed.contains(nick.toLowerCase())) {
                people.put(nick.toLowerCase(), p);
                // TODO: Send a message to new people to welcome the user and introduce the bot.
            }
        }
        p.notice(channel, type);
        return p;
    }

    public Person getPerson(String nick)
    {
        return people.get(nick.toLowerCase());
    }

    public Person getPerson(String nick, boolean create)
    {
        Person p = getPerson(nick);
        if (p == null && create) {
            p = notice(nick, null, NoticeType.UNSEEN);
        }
        return p;
    }

    public boolean isOperator(String nick)
    {
        Person person = getPerson(nick);
        return person != null && person.isOperator();
    }

    public void addOperator(String nick)
    {
        Person person = getPerson(nick);
        if (person != null)
            person.setOperator(true);
    }

    public boolean ignores(String nick)
    {
        Person p = people.get(nick);
        if (p == null)
            return false;
        return p.isIgnored();
    }

    public boolean isRemoved(String nick)
    {
        return removed.contains(nick.toLowerCase());
    }
    
    @SuppressWarnings("unchecked")
    public void loadState(SaveState state)
    {
        people = (HashMap<String, Person>)state.get("people");
        removed = (HashSet<String>)state.get("removed");

        if (people == null) people = new HashMap<String, Person>();
        if (removed == null) removed = new HashSet<String>();
    }

    public void saveState(SaveState state)
    {
        state.add("people", people);
        state.add("removed", removed);
    }

    public String getDescription(String topic, Language lang)
    {
        if (topic.equals(getModuleName().toLowerCase())) {
            return "Denna modul håller ordning på vilka som är aktiva i " +
                "kanalen.";
        }
        else if (topic.equals("!seen")) {
            return "!seen [channel] [nick [nick...]] --- Visar när användaren 'nick' senast var " +
                "aktiv i kanalen. Om inget nick anges visas alla personer " +
                "som varit aktiva i kanalen.";
        }
        else if (topic.equals("!info")) {
            return "!info [nick] --- Visa information om 'nick', eller om dig själv om inget 'nick' anges.";
        }
        else if (topic.equals("!score")) {
            return "!score --- Visar din poäng.";
        }
        else if (topic.equals("!highscore")) {
            return "!highscore --- Visar vilka i kanalen som har flest " +
                "poäng. Poäng kan erhållas genom att prata om rätt saker i " +
                "kanalen (och undvika fel saker), genom att vinna " +
                "tävlingar eller på annat sätt " +
                "visa att man är en bättre människa än genomsnittet.";
        }
        else if (topic.equals("!rename")) {
            return "!rename <new nick> --- Ändrar ditt nick i databasen.";
        }
        else if (topic.equals("!description")) {
            return "!description <text> --- Ange en personlig beskrivning av dig själv.";
        }
        else if (topic.equals("!greeting")) {
            return "!greeting <text> --- Ange en personlig hälsningsfras som kommer att användas när du går in i kanalen.";
        }
        else if (topic.equals("!alias")) {
            return "!alias [nick] [alias] --- Om ett 'nick' och ett 'alias' anges kommer dessa " +
                "namn att betraktas som samma person i fortsättningen. Om endast ett 'nick' anges " +
                "listas eventuella alias för detta. Om ingen parameter anges listas eventuella alias " +
                "för den som anropar.";
        }
        else if (topic.equals("!botop")) {
            return "!botop [remove] [nick] --- Ge 'nick' operatörsstatus för boten. " +
                "Om inget nick anges listas alla operatörer.";
        }
        else if (topic.equals("!trustee")) {
            return "!trustee [remove] [nick] --- Ge 'nick' trustee-status för boten. " +
                "Om inget nick anges listas alla trustees.";
        }
        else if (topic.equals("!reward")) {
            return "!reward <nick> [score] --- Ge nick poäng. Om score " +
                "anges ökas användarens poäng med så mycket, annars med " +
                "ett slumpmässigt värde mellan 20 och 50.";
        }
        else if (topic.equals("!punish")) {
            return "!punish <nick> [score] --- Bestraffa nick genom att " +
                "ta poäng. Om score anges minskas användarens poäng med så " +
                "mycket, annars med ett slumpmässigt värde mellan 20 och 50.";
        }
        else if (topic.equals("!ignore")) {
            return "!ignore [remove <nick>] [nick1] [nick2...] --- Låtsas som att 'nick' inte finns. " +
                "Flera nick kan anges för att ignorera flera personer. Om 'remove' anges slutar boten " +
                "att ignorera personen/personerna. Om inget 'nick' anges listas alla som ignoreras.";
        }
        else if (topic.equals("!remove")) {
            return "!remove <nick> --- Plocka bort 'nick' ur databasen. " +
                "Nytt nick för personen blir det första alias som finns " +
                "sparat. Om inget alias finns för 'nick' raderas all sparad " +
                "data om 'nick'.";
        }
        else
            return null;
    }

    public String getAuthor()
    {
        return "Jesper Wilhelmsson";
    }
}
