package ircbot.modules;

import java.util.*;
import ircbot.*;

public class Trivia extends AbstractIRCBotModule
{
    private HashMap<String, Game> games;
    private static ArrayList<Question> questions;

    private enum Topic { ENTERTAINMENT, MUSIC, MOVIES, HISTORY, MISC }

    public Trivia()
    {
        games = new HashMap<String, Game>();
        questions = new ArrayList<Question>();
        setupQuestions();
        addCommand(Modules.CommandType.PUBLIC, "!trivia");
    }

    public String getModuleName()
    {
        return "Trivial Pursuit";
    }

    public String action( IRCMessage message )
    {
        if (games.containsKey(message.getReplyTo())) {
            Game game = games.get(message.getReplyTo());
            return game.guess(message);
        } else {
            Game g = new Game(message.getReplyTo(), message.getParam(0));
            games.put(message.getReplyTo(), g);
            return g.question.question;
        }
    }

    public String parseMessage( IRCMessage message )
    {
        if (games.containsKey(message.getReplyTo())) {
            if (games.get(message.getReplyTo()).lastActive.getTime() > System.currentTimeMillis() - Time.MIN_5)
                return action( message );
        }

        return null;
    }

    public String getDescription(String topic, Language lang)
    {
        return "!trivia [topic | answer] --- startar ett nytt spel och ger dig en fråga. Om 'topic' anges kommer endast " +
        "frågor ur den kategorin. Tillgängliga kategorier är: " + Topic.values() + " " +
        "Du svarar på frågan genom att skriva !trivia <ditt svar>. Tänk på att försöka svara så " +
        "kortfattat som möjligt och utan att lägga på några onödiga tecken. Den som svarar rätt på " +
        "frågan får poäng.";
    }

    public String getAuthor()
    {
        return "Adam Viklund, Jesper Wilhelmsson";
    }

    private void setupQuestions()
    {
        questions.add(new Question(Topic.MOVIES, "Vem regisserade Pulp Fiction?", "quentin tarantino"));
        questions.add(new Question(Topic.MUSIC, "Vem skrev låtarna 'Keep on rockin in the free world' och 'Hey hey my my'?",
                                   "neil young"));
        questions.add(new Question(Topic.ENTERTAINMENT, "Vilket företag ligger från början bakom 'Command and Conquer'?",
                                   "westwood studios"));
        questions.add(new Question(Topic.MUSIC, "Vilken grupp skrev låtarna 'A day in the life' och 'For the benefit of mr. Kite'?",
                                   "the beatles"));
        questions.add(new Question(Topic.MUSIC, "Vilket år splittrades Refused? (format åååå)", "1998"));
        questions.add(new Question(Topic.MISC, "Vad står förkortningen FUBAR för?", "fucked up beyond all recognition"));
        questions.add(new Question(Topic.HISTORY, "Vilket år grundades Uppsala Universitet? (format åååå)", "1477"));
        questions.add(new Question(Topic.MISC, "I vilken grupp finns det medlemmar som General Knas och Sture Allen?",
                                   "svenska akademin"));
        questions.add(new Question(Topic.MUSIC, "Vad hette Metallicas basist 2006?", "robert trujillo"));
        questions.add(new Question(Topic.MUSIC, "Vad heter gitarristen som brukar spela med Ozzy Osbourne?", "zakk wylde"));
    }

    private class Game
    {
        private String owner;
        private Question question;
        private int number;
        private Date lastActive;
        private Topic topic;

        public Game(String owner, String topic)
        {
            this.owner = owner;
            this.topic = Topic.valueOf(topic); // TODO: If no category is given, any question will do.
            newQuestion();
            lastActive = new Date();
        }

        private void newQuestion()
        {
            number = Calc.rnd(questions.size());
            question = questions.get(number);
            int start = number++;
            if (number == questions.size())
                number = 0;
            while ((question.isAnswered() || question.topic != topic) && number != start) {
                if (number == questions.size())
                    number = 0;
                number++;
            }
        }

        public String guess(IRCMessage message)
        {
            if (message.getBotCommand() != null && message.getParams() == null) {
                lastActive = new Date();
                return question.question;
            }

            String answer;
            boolean forceAnswer = false;

            if (message.getBotCommand() != null) {
                answer = message.getParams().toLowerCase();
                forceAnswer = true;
            } else {
                answer = message.getMessage().toLowerCase();
            }

            if (answer.startsWith("bottenanna:")) {
                answer = answer.substring(10);
                forceAnswer = true;
            }

            if (question.isAnswered()) {
                if (answer.equals("ja")) {
                    newQuestion();
                    lastActive = new Date();
                    return question.question;
                } else if (answer.equals("nej")) {
                    games.remove(owner);
                    return "Trivia terminated.";
                }
            } else {
                Person player = message.getSender();
                float correctness = question.test2(answer);
                if (correctness > 0.95) {
                    player.incScore(5);
                    question.answered = true;
                    lastActive = new Date();
                    return player + ": Ja, rätt svar är '" + question.answer + "'. En fråga till? (ja/nej)";
                } else if (correctness > 0.8) {
                    player.incScore(2);
                    question.answered = true;
                    lastActive = new Date();
                    return player + ": Nära nog. Rätt svar är '" + question.answer + "'. En fråga till? (ja/nej)";
                } else if (correctness > 0.5 || forceAnswer) {
                    lastActive = new Date();
                    return player + ": Ditt svar var " + (int)(correctness * 100) + "% rätt.";
                }
            }
            return null;
        }
    }

    private class Question
    {
        String question;
        char[] answerSoundex;
        String answer;
        boolean answered;
        Topic topic;

        public Question(Topic category, String q, String a)
        {
            question = q;
            answer = a.toLowerCase();
            topic = category;
            answerSoundex = generateSoundex(answer);
            answered = false;
        }

        /** Assumes lower case text. */
        private char[] generateSoundex(String text)
        {
            String sndx = "";
            char ch = '\0';

            for (int i = 0; i < text.length(); i++) {
                boolean number = false;
                switch(text.charAt(i)) {
                case 'b':
                case 'p':
                case 'v':
                case 'w':
                case 'f':
                    ch = 'B';
                    break;
                case 'c':
                case 'g':
                case 'k':
                case 'q':
                case 's':
                case 'x':
                case 'z':
                    ch = 'C';
                    break;
                case 'd':
                case 't':
                    ch = 'D';
                    break;
                case 'm':
                case 'n':
                    ch = 'M';
                    break;
                case 'e':
                case 'ä':
                    ch = 'E';
                    break;
                case 'i':
                case 'j':
                    ch = 'I';
                    break;
                case 'o':
                case 'u':
                    ch = 'O';
                    break;
                case '1':
                case '2':
                case '3':
                case '4':
                case '5':
                case '6':
                case '7':
                case '8':
                case '9':
                case '0':
                    number = true;
                case 'l':
                case 'a':
                case 'y':
                case 'å':
                case 'ö':
                    ch = text.charAt(i);
                default:
                    break;
                }
                if (ch != '\0' && (number || sndx.length() == 0 || sndx.charAt(sndx.length() - 1) != ch))
                    sndx += ch;
            }
            return sndx.toCharArray();
        }

        public boolean isAnswered()
        {
            return answered;
        }

        public float test(String text)
        {
            char[] soundex = generateSoundex(text);
            int correct = 0;
            System.out.println("Soundex: " + new String(soundex) + "   Answer: " + new String(answerSoundex));
            int i = 0;
            int j = 0;
            boolean flip = true;
            while (i < soundex.length && j < answerSoundex.length) {
                if (soundex[i] == answerSoundex[j]) {
                    correct++;
                    i++; j++;
                } else {
                    if (flip)
                        i++;
                    else
                        j++;
                    flip = !flip;
                }
            }

            return (float)correct / answerSoundex.length;
        }

        public float test2(String text)
        {
            char[] soundex = generateSoundex(text);
            int[] subStrings = new int[answerSoundex.length];
            String answer = new String(answerSoundex);
            int longest = 0;

            System.out.println("Soundex: " + new String(soundex) + "   Answer: " + answer);

            for (int i = 0; i < subStrings.length; i++) {
                int startpos = 0;

                subStrings[i] = 0;
                while (subStrings.length - startpos > subStrings[i]) {
                    int match = 0;
                    int j = i;
                    int pos = text.indexOf(answer.substring(i), startpos);
                    startpos = pos;

                    while (text.charAt(pos++) == answer.charAt(j++))
                        match++;
                    if (match > subStrings[i]) {
                        subStrings[i] = match;
                        if (match > longest)
                            longest = match;
                    }
                }
            }

            return (float)longest / answerSoundex.length;
        }
    }
}
